@@ -1,8 +1,29 @@
-#define MOD_LISP_VERSION "2.2"
-#define HEADER_STR_LEN 100
+#define MOD_LISP_VERSION "2.33"
+#define HEADER_STR_LEN 500
 
 /*
-  Version 2.2 
+  Version 2.33
+  Added a couple of new headers like "Log-Notice" and so on. 
+  They are named like the corresponding log levels in httpd_log.h. 
+  The "default" log level (i.e. the one sent by just "Log") has not changed.
+  (contributed by Edi Weitz)
+
+  Version 2.32
+  Removed duplicate URL header sent to Lisp.
+  moved server-id header before the user http headers for security.
+  do not transmit any "end" header that could be sent by an malicious user
+  (Thanks to Robert Macomber for the security screening)
+
+  Version 2.31
+  Put back the correct handling of replies without known content length.
+  Reads only the missing number of bytes in the reply when the browser aborts the connection.
+
+  Version 2.3
+  Force Apache to read all the Lisp reply before eventually closing the socket or handling another request.
+  This avoids trying to write to a closed socket or having Lisp and Apache out of sync.
+  (contributed by Edi Weitz)
+
+  Version 2.2
   Allow more than one Set-Cookie
   Remaned the win32 dll to mod_lisp.dll
   
@@ -487,13 +508,13 @@ void CloseLispSocket(excfg *cfg, int Socket) // socket for WIN32
 #endif
 }
 
-int ForceGets(char *s, BUFF *BuffSocket)
+int ForceGets(char *s, BUFF *BuffSocket, int len)
 {
   int ret, i;
   
   for (i =0; i < 10; i++)
     {
-      ret = ap_bgets(s, HEADER_STR_LEN,  BuffSocket);
+      ret = ap_bgets(s, len,  BuffSocket);
       if (ret > 0)
 	return ret;
       ap_bsetflag(BuffSocket, B_RD, 1);
@@ -599,8 +620,6 @@ static int lisp_handler(request_rec *r)
 	  if (!strncmp(elts[i].key, "HTTP_", 5)) continue;
 	  if (!strcmp(elts[i].key, "REQUEST_URI"))
 	    ret = SendLispHeaderLine(BuffSocket, "url", elts[i].val);
-	  else if (!strcmp(elts[i].key, "REQUEST_URI"))
-	    ret = SendLispHeaderLine(BuffSocket, "url", elts[i].val);
 	  else if (!strcmp(elts[i].key, "CONTENT_TYPE"))
 	    ret = SendLispHeaderLine(BuffSocket, "content-type", elts[i].val);
 	  else if (!strcmp(elts[i].key, "CONTENT_LENGTH"))
@@ -629,6 +648,17 @@ static int lisp_handler(request_rec *r)
 	}
     }
   
+  /* Send this before client headers so ASSOC can be used to grab it
+   * without worrying about some joker sending a server-id header of
+   * his own. (Robert Macomber)*/
+  ret = SendLispHeaderLine(BuffSocket, "server-id", dcfg->LispServerId);
+  if (ret!=0)
+    {
+      ap_kill_timeout(r);
+      CloseLispSocket(dcfg, Socket);
+      return SERVER_ERROR;
+    }
+
   if (r->headers_in) 
     {
       array_header *hdr_arr = ap_table_elts(r->headers_in);
@@ -638,8 +668,10 @@ static int lisp_handler(request_rec *r)
       for (i = 0; i < hdr_arr->nelts; ++i) 
 	{
 	  if (!elts[i].key) continue;
-	  ret = SendLispHeaderLine(BuffSocket, elts[i].key, elts[i].val);
-	  if (ret!=0) 
+	  ret = SendLispHeaderLine(BuffSocket, 
+				   !strcmp(elts[i].key, "end")?"end-header":elts[i].key, 
+				   elts[i].val);
+	  if (ret!=0)
 	    {
 	      ap_kill_timeout(r);
 	      CloseLispSocket(dcfg, Socket);
@@ -649,14 +681,6 @@ static int lisp_handler(request_rec *r)
 	}
     }
   
-  ret = SendLispHeaderLine(BuffSocket, "server-id", dcfg->LispServerId);
-  if (ret!=0)
-    {
-      ap_kill_timeout(r);
-      CloseLispSocket(dcfg, Socket);
-      return SERVER_ERROR;
-    }
-
   if (SendLispString(BuffSocket, "end\n") == -1 || FlushLispBuffSocket(BuffSocket) == -1) 
     {
       ap_kill_timeout(r);
@@ -699,7 +723,7 @@ static int lisp_handler(request_rec *r)
   /* Receive the response from Lisp */
   ap_hard_timeout("lisp-read", r);
 
-  while (ForceGets(Header, (BUFF *) BuffSocket) >= 0)
+  while (ForceGets(Header, (BUFF *) BuffSocket, HEADER_STR_LEN) >= 0)
     {
       int l;
      l = strlen(Header);
@@ -747,13 +771,41 @@ static int lisp_handler(request_rec *r)
 	{
 	  KeepSocket = atoi(Value);
 	}
+      else if (!strcmp(Header, "Log-Emerg"))
+	{
+	  ap_log_error("mod_lisp", 0, APLOG_EMERG|APLOG_NOERRNO, r->server, "%s", Value);
+	}
+      else if (!strcmp(Header, "Log-Alert"))
+	{
+	  ap_log_error("mod_lisp", 0, APLOG_ALERT|APLOG_NOERRNO, r->server, "%s", Value);
+	}
+      else if (!strcmp(Header, "Log-Crit"))
+	{
+	  ap_log_error("mod_lisp", 0, APLOG_CRIT|APLOG_NOERRNO, r->server, "%s", Value);
+	}
       else if (!strcmp(Header, "Log-Error"))
 	{
-	  ap_log_error("mod_lisp", 0, APLOG_ERR, r->server, Value);
+	  ap_log_error("mod_lisp", 0, APLOG_ERR|APLOG_NOERRNO, r->server, "%s", Value);
+	}
+      else if (!strcmp(Header, "Log-Warning"))
+	{
+	  ap_log_error("mod_lisp", 0, APLOG_WARNING|APLOG_NOERRNO, r->server, "%s", Value);
+	}
+      else if (!strcmp(Header, "Log-Notice"))
+	{
+	  ap_log_error("mod_lisp", 0, APLOG_NOTICE|APLOG_NOERRNO, r->server, "%s", Value);
+	}
+      else if (!strcmp(Header, "Log-Info"))
+	{
+	  ap_log_error("mod_lisp", 0, APLOG_INFO|APLOG_NOERRNO, r->server, "%s", Value);
+	}
+      else if (!strcmp(Header, "Log-Debug"))
+	{
+	  ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, r->server, "%s", Value);
 	}
       else if (!strcmp(Header, "Log"))
 	{
-	  ap_log_error("mod_lisp", 0, APLOG_ERR, r->server, Value);
+	  ap_log_error("mod_lisp", 0, APLOG_ERR|APLOG_NOERRNO, r->server, "%s", Value);
 	}
       else if (!strcmp(Header, "Note"))
 	{
@@ -774,8 +826,26 @@ static int lisp_handler(request_rec *r)
   
   /* Send headers and data collected (if this was not a "header only" req. */
   ap_send_http_header(r);
-  if (!r->header_only && ContentLength > 0) 
-    ap_send_fb_length(BuffSocket, r, ContentLength);
+  if (!r->header_only)
+    if (ContentLength > 0)
+      {
+	long ReadLength = ap_send_fb_length(BuffSocket, r, ContentLength);
+	if (ReadLength < ContentLength || r->connection->aborted)
+	  {
+	    char buffer[HUGE_STRING_LEN];
+	    ContentLength -= ReadLength;
+	    do
+	      {
+		ReadLength = ForceGets(buffer, (BUFF *) BuffSocket, 
+				       HUGE_STRING_LEN > ContentLength ? ContentLength : HUGE_STRING_LEN);
+		ContentLength -= ReadLength;
+	      }
+	    while (ReadLength > 0 && ContentLength > 0);
+	  }
+      }
+      else
+	if (ContentLength == -1)
+	  ap_send_fb(BuffSocket, r);
 
 #ifdef WIN32
   KeepSocket = 0;
